@@ -12,7 +12,6 @@ import type {BrowserContext} from '../api/BrowserContext.js';
 import {CDPSessionEvent, type CDPSession} from '../api/CDPSession.js';
 import type {ElementHandle} from '../api/ElementHandle.js';
 import type {Frame, WaitForOptions} from '../api/Frame.js';
-import type {HTTPRequest} from '../api/HTTPRequest.js';
 import type {HTTPResponse} from '../api/HTTPResponse.js';
 import type {JSHandle} from '../api/JSHandle.js';
 import type {Credentials} from '../api/Page.js';
@@ -35,8 +34,10 @@ import type {
   Cookie,
   DeleteCookiesRequest,
   CookieParam,
+  CookiePartitionKey,
 } from '../common/Cookie.js';
 import {TargetCloseError} from '../common/Errors.js';
+import {EventEmitter} from '../common/EventEmitter.js';
 import {FileChooser} from '../common/FileChooser.js';
 import {NetworkManagerEvent} from '../common/NetworkManagerEvents.js';
 import type {PDFOptions} from '../common/PDFOptions.js';
@@ -44,7 +45,7 @@ import type {BindingPayload, HandleFor} from '../common/types.js';
 import {
   debugError,
   evaluationString,
-  getReadableAsBuffer,
+  getReadableAsTypedArray,
   getReadableFromProtocolStream,
   parsePDFOptions,
   timeout,
@@ -56,7 +57,6 @@ import {Deferred} from '../util/Deferred.js';
 import {AsyncDisposableStack} from '../util/disposable.js';
 import {isErrorLike} from '../util/ErrorLike.js';
 
-import {Accessibility} from './Accessibility.js';
 import {Binding} from './Binding.js';
 import {CdpCDPSession} from './CDPSession.js';
 import {isTargetClosedError} from './Connection.js';
@@ -64,7 +64,6 @@ import {Coverage} from './Coverage.js';
 import type {DeviceRequestPrompt} from './DeviceRequestPrompt.js';
 import {CdpDialog} from './Dialog.js';
 import {EmulationManager} from './EmulationManager.js';
-import {FirefoxTargetManager} from './FirefoxTargetManager.js';
 import type {CdpFrame} from './Frame.js';
 import {FrameManager} from './FrameManager.js';
 import {FrameManagerEvent} from './FrameManagerEvents.js';
@@ -74,8 +73,8 @@ import {MAIN_WORLD} from './IsolatedWorlds.js';
 import {releaseObject} from './JSHandle.js';
 import type {NetworkConditions} from './NetworkManager.js';
 import type {CdpTarget} from './Target.js';
+import {TargetManagerEvent} from './TargetManageEvents.js';
 import type {TargetManager} from './TargetManager.js';
-import {TargetManagerEvent} from './TargetManager.js';
 import {Tracing} from './Tracing.js';
 import {
   createClientError,
@@ -100,7 +99,7 @@ export class CdpPage extends Page {
   static async _create(
     client: CDPSession,
     target: CdpTarget,
-    defaultViewport: Viewport | null
+    defaultViewport: Viewport | null,
   ): Promise<CdpPage> {
     const page = new CdpPage(client, target);
     await page.#initialize();
@@ -128,7 +127,6 @@ export class CdpPage extends Page {
   #keyboard: CdpKeyboard;
   #mouse: CdpMouse;
   #touchscreen: CdpTouchscreen;
-  #accessibility: Accessibility;
   #frameManager: FrameManager;
   #emulationManager: EmulationManager;
   #tracing: Tracing;
@@ -142,89 +140,6 @@ export class CdpPage extends Page {
   #serviceWorkerBypassed = false;
   #userDragInterceptionEnabled = false;
 
-  readonly #frameManagerHandlers = [
-    [
-      FrameManagerEvent.FrameAttached,
-      (frame: CdpFrame) => {
-        this.emit(PageEvent.FrameAttached, frame);
-      },
-    ],
-    [
-      FrameManagerEvent.FrameDetached,
-      (frame: CdpFrame) => {
-        this.emit(PageEvent.FrameDetached, frame);
-      },
-    ],
-    [
-      FrameManagerEvent.FrameNavigated,
-      (frame: CdpFrame) => {
-        this.emit(PageEvent.FrameNavigated, frame);
-      },
-    ],
-  ] as const;
-
-  readonly #networkManagerHandlers = [
-    [
-      NetworkManagerEvent.Request,
-      (request: HTTPRequest) => {
-        this.emit(PageEvent.Request, request);
-      },
-    ],
-    [
-      NetworkManagerEvent.RequestServedFromCache,
-      (request: HTTPRequest) => {
-        this.emit(PageEvent.RequestServedFromCache, request);
-      },
-    ],
-    [
-      NetworkManagerEvent.Response,
-      (response: HTTPResponse) => {
-        this.emit(PageEvent.Response, response);
-      },
-    ],
-    [
-      NetworkManagerEvent.RequestFailed,
-      (request: HTTPRequest) => {
-        this.emit(PageEvent.RequestFailed, request);
-      },
-    ],
-    [
-      NetworkManagerEvent.RequestFinished,
-      (request: HTTPRequest) => {
-        this.emit(PageEvent.RequestFinished, request);
-      },
-    ],
-  ] as const;
-
-  readonly #sessionHandlers = [
-    [
-      CDPSessionEvent.Disconnected,
-      () => {
-        this.#sessionCloseDeferred.reject(
-          new TargetCloseError('Target closed')
-        );
-      },
-    ],
-    [
-      'Page.domContentEventFired',
-      () => {
-        return this.emit(PageEvent.DOMContentLoaded, undefined);
-      },
-    ],
-    [
-      'Page.loadEventFired',
-      () => {
-        return this.emit(PageEvent.Load, undefined);
-      },
-    ],
-    ['Page.javascriptDialogOpening', this.#onDialog.bind(this)],
-    ['Runtime.exceptionThrown', this.#handleException.bind(this)],
-    ['Inspector.targetCrashed', this.#onTargetCrashed.bind(this)],
-    ['Performance.metrics', this.#emitMetrics.bind(this)],
-    ['Log.entryAdded', this.#onLogEntryAdded.bind(this)],
-    ['Page.fileChooserOpened', this.#onFileChooser.bind(this)],
-  ] as const;
-
   constructor(client: CDPSession, target: CdpTarget) {
     super();
     this.#primaryTargetClient = client;
@@ -237,55 +152,70 @@ export class CdpPage extends Page {
     this.#keyboard = new CdpKeyboard(client);
     this.#mouse = new CdpMouse(client, this.#keyboard);
     this.#touchscreen = new CdpTouchscreen(client, this.#keyboard);
-    this.#accessibility = new Accessibility(client);
     this.#frameManager = new FrameManager(client, this, this._timeoutSettings);
     this.#emulationManager = new EmulationManager(client);
     this.#tracing = new Tracing(client);
     this.#coverage = new Coverage(client);
     this.#viewport = null;
 
-    for (const [eventName, handler] of this.#frameManagerHandlers) {
-      this.#frameManager.on(eventName, handler);
-    }
-
-    this.#frameManager.on(
+    const frameManagerEmitter = new EventEmitter(this.#frameManager);
+    frameManagerEmitter.on(FrameManagerEvent.FrameAttached, frame => {
+      this.emit(PageEvent.FrameAttached, frame);
+    });
+    frameManagerEmitter.on(FrameManagerEvent.FrameDetached, frame => {
+      this.emit(PageEvent.FrameDetached, frame);
+    });
+    frameManagerEmitter.on(FrameManagerEvent.FrameNavigated, frame => {
+      this.emit(PageEvent.FrameNavigated, frame);
+    });
+    frameManagerEmitter.on(
       FrameManagerEvent.ConsoleApiCalled,
-      ([world, event]: [
-        IsolatedWorld,
-        Protocol.Runtime.ConsoleAPICalledEvent,
-      ]) => {
+      ([world, event]) => {
         this.#onConsoleAPI(world, event);
-      }
+      },
     );
-
-    this.#frameManager.on(
+    frameManagerEmitter.on(
       FrameManagerEvent.BindingCalled,
-      ([world, event]: [
-        IsolatedWorld,
-        Protocol.Runtime.BindingCalledEvent,
-      ]) => {
+      ([world, event]) => {
         void this.#onBindingCalled(world, event);
-      }
+      },
     );
 
-    for (const [eventName, handler] of this.#networkManagerHandlers) {
-      // TODO: Remove any.
-      this.#frameManager.networkManager.on(eventName, handler as any);
-    }
+    const networkManagerEmitter = new EventEmitter(
+      this.#frameManager.networkManager,
+    );
+    networkManagerEmitter.on(NetworkManagerEvent.Request, request => {
+      this.emit(PageEvent.Request, request);
+    });
+    networkManagerEmitter.on(
+      NetworkManagerEvent.RequestServedFromCache,
+      request => {
+        this.emit(PageEvent.RequestServedFromCache, request!);
+      },
+    );
+    networkManagerEmitter.on(NetworkManagerEvent.Response, response => {
+      this.emit(PageEvent.Response, response);
+    });
+    networkManagerEmitter.on(NetworkManagerEvent.RequestFailed, request => {
+      this.emit(PageEvent.RequestFailed, request);
+    });
+    networkManagerEmitter.on(NetworkManagerEvent.RequestFinished, request => {
+      this.emit(PageEvent.RequestFinished, request);
+    });
 
     this.#tabTargetClient.on(
       CDPSessionEvent.Swapped,
-      this.#onActivation.bind(this)
+      this.#onActivation.bind(this),
     );
 
     this.#tabTargetClient.on(
       CDPSessionEvent.Ready,
-      this.#onSecondaryTarget.bind(this)
+      this.#onSecondaryTarget.bind(this),
     );
 
     this.#targetManager.on(
       TargetManagerEvent.TargetGone,
-      this.#onDetachedFromTarget
+      this.#onDetachedFromTarget,
     );
 
     this.#tabTarget._isClosedDeferred
@@ -293,7 +223,7 @@ export class CdpPage extends Page {
       .then(() => {
         this.#targetManager.off(
           TargetManagerEvent.TargetGone,
-          this.#onDetachedFromTarget
+          this.#onDetachedFromTarget,
         );
 
         this.emit(PageEvent.Close, undefined);
@@ -302,20 +232,41 @@ export class CdpPage extends Page {
       .catch(debugError);
 
     this.#setupPrimaryTargetListeners();
+    this.#attachExistingTargets();
+  }
+
+  #attachExistingTargets(): void {
+    const queue = [];
+    for (const childTarget of this.#targetManager.getChildTargets(
+      this.#primaryTarget,
+    )) {
+      queue.push(childTarget);
+    }
+    let idx = 0;
+    while (idx < queue.length) {
+      const next = queue[idx] as CdpTarget;
+      idx++;
+      const session = next._session();
+      if (session) {
+        this.#onAttachedToTarget(session);
+      }
+      for (const childTarget of this.#targetManager.getChildTargets(next)) {
+        queue.push(childTarget);
+      }
+    }
   }
 
   async #onActivation(newSession: CDPSession): Promise<void> {
     this.#primaryTargetClient = newSession;
     assert(
       this.#primaryTargetClient instanceof CdpCDPSession,
-      'CDPSession is not instance of CDPSessionImpl'
+      'CDPSession is not instance of CDPSessionImpl',
     );
     this.#primaryTarget = this.#primaryTargetClient._target();
     assert(this.#primaryTarget, 'Missing target on swap');
     this.#keyboard.updateClient(newSession);
     this.#mouse.updateClient(newSession);
     this.#touchscreen.updateClient(newSession);
-    this.#accessibility.updateClient(newSession);
     this.#emulationManager.updateClient(newSession);
     this.#tracing.updateClient(newSession);
     this.#coverage.updateClient(newSession);
@@ -339,15 +290,29 @@ export class CdpPage extends Page {
    * during a navigation to a prerended page.
    */
   #setupPrimaryTargetListeners() {
-    this.#primaryTargetClient.on(
-      CDPSessionEvent.Ready,
-      this.#onAttachedToTarget
+    const clientEmitter = new EventEmitter(this.#primaryTargetClient);
+    clientEmitter.on(CDPSessionEvent.Ready, this.#onAttachedToTarget);
+    clientEmitter.on(CDPSessionEvent.Disconnected, () => {
+      this.#sessionCloseDeferred.reject(new TargetCloseError('Target closed'));
+    });
+    clientEmitter.on('Page.domContentEventFired', () => {
+      this.emit(PageEvent.DOMContentLoaded, undefined);
+    });
+    clientEmitter.on('Page.loadEventFired', () => {
+      this.emit(PageEvent.Load, undefined);
+    });
+    clientEmitter.on('Page.javascriptDialogOpening', this.#onDialog.bind(this));
+    clientEmitter.on(
+      'Runtime.exceptionThrown',
+      this.#handleException.bind(this),
     );
-
-    for (const [eventName, handler] of this.#sessionHandlers) {
-      // TODO: Remove any.
-      this.#primaryTargetClient.on(eventName, handler as any);
-    }
+    clientEmitter.on(
+      'Inspector.targetCrashed',
+      this.#onTargetCrashed.bind(this),
+    );
+    clientEmitter.on('Performance.metrics', this.#emitMetrics.bind(this));
+    clientEmitter.on('Log.entryAdded', this.#onLogEntryAdded.bind(this));
+    clientEmitter.on('Page.fileChooserOpened', this.#onFileChooser.bind(this));
   }
 
   #onDetachedFromTarget = (target: CdpTarget) => {
@@ -370,7 +335,7 @@ export class CdpPage extends Page {
         session._target()._targetId,
         session._target().type(),
         this.#addConsoleMessage.bind(this),
-        this.#handleException.bind(this)
+        this.#handleException.bind(this),
       );
       this.#workers.set(session.id(), worker);
       this.emit(PageEvent.WorkerCreated, worker);
@@ -395,7 +360,7 @@ export class CdpPage extends Page {
   }
 
   async #onFileChooser(
-    event: Protocol.Page.FileChooserOpenedEvent
+    event: Protocol.Page.FileChooserOpenedEvent,
   ): Promise<void> {
     if (!this.#fileChooserDeferreds.size) {
       return;
@@ -406,7 +371,7 @@ export class CdpPage extends Page {
 
     // This is guaranteed to be an HTMLInputElement handle by the event.
     using handle = (await frame.worlds[MAIN_WORLD].adoptBackendNode(
-      event.backendNodeId
+      event.backendNodeId,
     )) as ElementHandle<HTMLInputElement>;
 
     const fileChooser = new FileChooser(handle.move(), event);
@@ -433,7 +398,7 @@ export class CdpPage extends Page {
   }
 
   override async waitForFileChooser(
-    options: WaitTimeoutOptions = {}
+    options: WaitTimeoutOptions = {},
   ): Promise<FileChooser> {
     const needsEnable = this.#fileChooserDeferreds.size === 0;
     const {timeout = this._timeoutSettings.timeout()} = options;
@@ -441,6 +406,17 @@ export class CdpPage extends Page {
       message: `Waiting for \`FileChooser\` failed: ${timeout}ms exceeded`,
       timeout,
     });
+
+    if (options.signal) {
+      options.signal.addEventListener(
+        'abort',
+        () => {
+          deferred.reject(options.signal?.reason);
+        },
+        {once: true},
+      );
+    }
+
     this.#fileChooserDeferreds.add(deferred);
     let enablePromise: Promise<void> | undefined;
     if (needsEnable) {
@@ -448,7 +424,7 @@ export class CdpPage extends Page {
         'Page.setInterceptFileChooserDialog',
         {
           enabled: true,
-        }
+        },
       );
     }
     try {
@@ -497,8 +473,8 @@ export class CdpPage extends Page {
           convertConsoleMessageLevel(level),
           text,
           [],
-          [{url, lineNumber}]
-        )
+          [{url, lineNumber}],
+        ),
       );
     }
   }
@@ -523,10 +499,6 @@ export class CdpPage extends Page {
     return this.#tracing;
   }
 
-  override get accessibility(): Accessibility {
-    return this.#accessibility;
-  }
-
   override frames(): Frame[] {
     return this.#frameManager.frames();
   }
@@ -537,7 +509,7 @@ export class CdpPage extends Page {
 
   override async setRequestInterception(value: boolean): Promise<void> {
     return await this.#frameManager.networkManager.setRequestInterception(
-      value
+      value,
     );
   }
 
@@ -545,7 +517,7 @@ export class CdpPage extends Page {
     this.#serviceWorkerBypassed = bypass;
     return await this.#primaryTargetClient.send(
       'Network.setBypassServiceWorker',
-      {bypass}
+      {bypass},
     );
   }
 
@@ -561,10 +533,10 @@ export class CdpPage extends Page {
   }
 
   override async emulateNetworkConditions(
-    networkConditions: NetworkConditions | null
+    networkConditions: NetworkConditions | null,
   ): Promise<void> {
     return await this.#frameManager.networkManager.emulateNetworkConditions(
-      networkConditions
+      networkConditions,
     );
   }
 
@@ -580,19 +552,23 @@ export class CdpPage extends Page {
     return this._timeoutSettings.timeout();
   }
 
+  override getDefaultNavigationTimeout(): number {
+    return this._timeoutSettings.navigationTimeout();
+  }
+
   override async queryObjects<Prototype>(
-    prototypeHandle: JSHandle<Prototype>
+    prototypeHandle: JSHandle<Prototype>,
   ): Promise<JSHandle<Prototype[]>> {
     assert(!prototypeHandle.disposed, 'Prototype JSHandle is disposed!');
     assert(
       prototypeHandle.id,
-      'Prototype JSHandle must not be referencing primitive value'
+      'Prototype JSHandle must not be referencing primitive value',
     );
     const response = await this.mainFrame().client.send(
       'Runtime.queryObjects',
       {
         prototypeObjectId: prototypeHandle.id,
-      }
+      },
     );
     return this.mainFrame()
       .mainRealm()
@@ -608,14 +584,23 @@ export class CdpPage extends Page {
 
     const unsupportedCookieAttributes = ['sourcePort'];
     const filterUnsupportedAttributes = (
-      cookie: Protocol.Network.Cookie
+      cookie: Protocol.Network.Cookie,
     ): Protocol.Network.Cookie => {
       for (const attr of unsupportedCookieAttributes) {
         delete (cookie as unknown as Record<string, unknown>)[attr];
       }
       return cookie;
     };
-    return originalCookies.map(filterUnsupportedAttributes);
+    return originalCookies.map(filterUnsupportedAttributes).map(cookie => {
+      return {
+        ...cookie,
+        // TODO: a breaking change is needed in Puppeteer types to support other
+        // partition keys.
+        partitionKey: cookie.partitionKey
+          ? cookie.partitionKey.topLevelSite
+          : undefined,
+      };
+    });
   }
 
   override async deleteCookie(
@@ -623,11 +608,27 @@ export class CdpPage extends Page {
   ): Promise<void> {
     const pageURL = this.url();
     for (const cookie of cookies) {
-      const item = Object.assign({}, cookie);
+      const item = {
+        ...cookie,
+        partitionKey: convertCookiesPartitionKeyFromPuppeteerToCdp(
+          cookie.partitionKey,
+        ),
+      };
       if (!cookie.url && pageURL.startsWith('http')) {
         item.url = pageURL;
       }
       await this.#primaryTargetClient.send('Network.deleteCookies', item);
+      if (pageURL.startsWith('http') && !item.partitionKey) {
+        const url = new URL(pageURL);
+        // Delete also cookies from the page's partition.
+        await this.#primaryTargetClient.send('Network.deleteCookies', {
+          ...item,
+          partitionKey: {
+            topLevelSite: url.origin.replace(`:${url.port}`, ''),
+            hasCrossSiteAncestor: false,
+          },
+        });
+      }
     }
   }
 
@@ -641,131 +642,103 @@ export class CdpPage extends Page {
       }
       assert(
         item.url !== 'about:blank',
-        `Blank page can not have cookie "${item.name}"`
+        `Blank page can not have cookie "${item.name}"`,
       );
       assert(
         !String.prototype.startsWith.call(item.url || '', 'data:'),
-        `Data URL page can not have cookie "${item.name}"`
+        `Data URL page can not have cookie "${item.name}"`,
       );
       return item;
     });
     await this.deleteCookie(...items);
     if (items.length) {
       await this.#primaryTargetClient.send('Network.setCookies', {
-        cookies: items,
+        cookies: items.map(cookieParam => {
+          return {
+            ...cookieParam,
+            partitionKey: convertCookiesPartitionKeyFromPuppeteerToCdp(
+              cookieParam.partitionKey,
+            ),
+          };
+        }),
       });
     }
   }
 
   override async exposeFunction(
     name: string,
-    pptrFunction: Function | {default: Function}
+    // eslint-disable-next-line @typescript-eslint/no-unsafe-function-type
+    pptrFunction: Function | {default: Function},
   ): Promise<void> {
     if (this.#bindings.has(name)) {
       throw new Error(
-        `Failed to add page binding with name ${name}: window['${name}'] already exists!`
+        `Failed to add page binding with name ${name}: window['${name}'] already exists!`,
       );
     }
-
+    const source = pageBindingInitString('exposedFun', name);
     let binding: Binding;
     switch (typeof pptrFunction) {
       case 'function':
         binding = new Binding(
           name,
-          pptrFunction as (...args: unknown[]) => unknown
+          pptrFunction as (...args: unknown[]) => unknown,
+          source,
         );
         break;
       default:
         binding = new Binding(
           name,
-          pptrFunction.default as (...args: unknown[]) => unknown
+          pptrFunction.default as (...args: unknown[]) => unknown,
+          source,
         );
         break;
     }
-
     this.#bindings.set(name, binding);
-
-    const expression = pageBindingInitString('exposedFun', name);
-    await this.#primaryTargetClient.send('Runtime.addBinding', {name});
-    // TODO: investigate this as it appears to only apply to the main frame and
-    // local subframes instead of the entire frame tree (including future
-    // frame).
-    const {identifier} = await this.#primaryTargetClient.send(
-      'Page.addScriptToEvaluateOnNewDocument',
-      {
-        source: expression,
-      }
-    );
-
+    const [{identifier}] = await Promise.all([
+      this.#frameManager.evaluateOnNewDocument(source),
+      this.#frameManager.addExposedFunctionBinding(binding),
+    ]);
     this.#exposedFunctions.set(name, identifier);
-
-    await Promise.all(
-      this.frames().map(frame => {
-        // If a frame has not started loading, it might never start. Rely on
-        // addScriptToEvaluateOnNewDocument in that case.
-        if (frame !== this.mainFrame() && !frame._hasStartedLoading) {
-          return;
-        }
-        return frame.evaluate(expression).catch(debugError);
-      })
-    );
   }
 
   override async removeExposedFunction(name: string): Promise<void> {
-    const exposedFun = this.#exposedFunctions.get(name);
-    if (!exposedFun) {
-      throw new Error(
-        `Failed to remove page binding with name ${name}: window['${name}'] does not exists!`
-      );
+    const exposedFunctionId = this.#exposedFunctions.get(name);
+    if (!exposedFunctionId) {
+      throw new Error(`Function with name "${name}" does not exist`);
     }
-
-    await this.#primaryTargetClient.send('Runtime.removeBinding', {name});
-    await this.removeScriptToEvaluateOnNewDocument(exposedFun);
-
-    await Promise.all(
-      this.frames().map(frame => {
-        // If a frame has not started loading, it might never start. Rely on
-        // addScriptToEvaluateOnNewDocument in that case.
-        if (frame !== this.mainFrame() && !frame._hasStartedLoading) {
-          return;
-        }
-        return frame
-          .evaluate(name => {
-            // Removes the dangling Puppeteer binding wrapper.
-            // @ts-expect-error: In a different context.
-            globalThis[name] = undefined;
-          }, name)
-          .catch(debugError);
-      })
-    );
-
+    // #bindings must be updated together with #exposedFunctions.
+    const binding = this.#bindings.get(name)!;
     this.#exposedFunctions.delete(name);
     this.#bindings.delete(name);
+    await Promise.all([
+      this.#frameManager.removeScriptToEvaluateOnNewDocument(exposedFunctionId),
+      this.#frameManager.removeExposedFunctionBinding(binding),
+    ]);
   }
 
-  override async authenticate(credentials: Credentials): Promise<void> {
+  override async authenticate(credentials: Credentials | null): Promise<void> {
     return await this.#frameManager.networkManager.authenticate(credentials);
   }
 
   override async setExtraHTTPHeaders(
-    headers: Record<string, string>
+    headers: Record<string, string>,
   ): Promise<void> {
     return await this.#frameManager.networkManager.setExtraHTTPHeaders(headers);
   }
 
   override async setUserAgent(
     userAgent: string,
-    userAgentMetadata?: Protocol.Emulation.UserAgentMetadata
+    userAgentMetadata?: Protocol.Emulation.UserAgentMetadata,
   ): Promise<void> {
     return await this.#frameManager.networkManager.setUserAgent(
       userAgent,
-      userAgentMetadata
+      userAgentMetadata,
     );
   }
 
   override async metrics(): Promise<Metrics> {
     const response = await this.#primaryTargetClient.send(
-      'Performance.getMetrics'
+      'Performance.getMetrics',
     );
     return this.#buildMetricsObject(response.metrics);
   }
@@ -793,13 +766,13 @@ export class CdpPage extends Page {
   #handleException(exception: Protocol.Runtime.ExceptionThrownEvent): void {
     this.emit(
       PageEvent.PageError,
-      createClientError(exception.exceptionDetails)
+      createClientError(exception.exceptionDetails),
     );
   }
 
   #onConsoleAPI(
     world: IsolatedWorld,
-    event: Protocol.Runtime.ConsoleAPICalledEvent
+    event: Protocol.Runtime.ConsoleAPICalledEvent,
   ): void {
     const values = event.args.map(arg => {
       return world.createCdpHandle(arg);
@@ -807,13 +780,13 @@ export class CdpPage extends Page {
     this.#addConsoleMessage(
       convertConsoleMessageLevel(event.type),
       values,
-      event.stackTrace
+      event.stackTrace,
     );
   }
 
   async #onBindingCalled(
     world: IsolatedWorld,
-    event: Protocol.Runtime.BindingCalledEvent
+    event: Protocol.Runtime.BindingCalledEvent,
   ): Promise<void> {
     let payload: BindingPayload;
     try {
@@ -840,7 +813,7 @@ export class CdpPage extends Page {
   #addConsoleMessage(
     eventType: string,
     args: JSHandle[],
-    stackTrace?: Protocol.Runtime.StackTrace
+    stackTrace?: Protocol.Runtime.StackTrace,
   ): void {
     if (!this.listenerCount(PageEvent.Console)) {
       args.forEach(arg => {
@@ -873,7 +846,7 @@ export class CdpPage extends Page {
       convertConsoleMessageLevel(eventType),
       textTokens.join(' '),
       args,
-      stackTraceLocations
+      stackTraceLocations,
     );
     this.emit(PageEvent.Console, message);
   }
@@ -884,13 +857,13 @@ export class CdpPage extends Page {
       this.#primaryTargetClient,
       type,
       event.message,
-      event.defaultPrompt
+      event.defaultPrompt,
     );
     this.emit(PageEvent.Dialog, dialog);
   }
 
   override async reload(
-    options?: WaitForOptions
+    options?: WaitForOptions,
   ): Promise<HTTPResponse | null> {
     const [result] = await Promise.all([
       this.waitForNavigation({
@@ -908,23 +881,23 @@ export class CdpPage extends Page {
   }
 
   override async goBack(
-    options: WaitForOptions = {}
+    options: WaitForOptions = {},
   ): Promise<HTTPResponse | null> {
     return await this.#go(-1, options);
   }
 
   override async goForward(
-    options: WaitForOptions = {}
+    options: WaitForOptions = {},
   ): Promise<HTTPResponse | null> {
     return await this.#go(+1, options);
   }
 
   async #go(
     delta: number,
-    options: WaitForOptions
+    options: WaitForOptions,
   ): Promise<HTTPResponse | null> {
     const history = await this.#primaryTargetClient.send(
-      'Page.getNavigationHistory'
+      'Page.getNavigationHistory',
     );
     const entry = history.entries[history.currentIndex + delta];
     if (!entry) {
@@ -960,7 +933,7 @@ export class CdpPage extends Page {
   }
 
   override async emulateMediaFeatures(
-    features?: MediaFeature[]
+    features?: MediaFeature[],
   ): Promise<void> {
     return await this.#emulationManager.emulateMediaFeatures(features);
   }
@@ -977,12 +950,12 @@ export class CdpPage extends Page {
   }
 
   override async emulateVisionDeficiency(
-    type?: Protocol.Emulation.SetEmulatedVisionDeficiencyRequest['type']
+    type?: Protocol.Emulation.SetEmulatedVisionDeficiencyRequest['type'],
   ): Promise<void> {
     return await this.#emulationManager.emulateVisionDeficiency(type);
   }
 
-  override async setViewport(viewport: Viewport): Promise<void> {
+  override async setViewport(viewport: Viewport | null): Promise<void> {
     const needsReload = await this.#emulationManager.emulateViewport(viewport);
     this.#viewport = viewport;
     if (needsReload) {
@@ -1002,24 +975,14 @@ export class CdpPage extends Page {
     ...args: Params
   ): Promise<NewDocumentScriptEvaluation> {
     const source = evaluationString(pageFunction, ...args);
-    const {identifier} = await this.#primaryTargetClient.send(
-      'Page.addScriptToEvaluateOnNewDocument',
-      {
-        source,
-      }
-    );
-
-    return {identifier};
+    return await this.#frameManager.evaluateOnNewDocument(source);
   }
 
   override async removeScriptToEvaluateOnNewDocument(
-    identifier: string
+    identifier: string,
   ): Promise<void> {
-    await this.#primaryTargetClient.send(
-      'Page.removeScriptToEvaluateOnNewDocument',
-      {
-        identifier,
-      }
+    return await this.#frameManager.removeScriptToEvaluateOnNewDocument(
+      identifier,
     );
   }
 
@@ -1028,7 +991,7 @@ export class CdpPage extends Page {
   }
 
   override async _screenshot(
-    options: Readonly<ScreenshotOptions>
+    options: Readonly<ScreenshotOptions>,
   ): Promise<string> {
     const {
       fromSurface,
@@ -1040,12 +1003,8 @@ export class CdpPage extends Page {
       captureBeyondViewport,
     } = options;
 
-    const isFirefox =
-      this.target()._targetManager() instanceof FirefoxTargetManager;
-
     await using stack = new AsyncDisposableStack();
-    // Firefox omits background by default; it's not configurable.
-    if (!isFirefox && omitBackground && (type === 'png' || type === 'webp')) {
+    if (omitBackground && (type === 'png' || type === 'webp')) {
       await this.#emulationManager.setTransparentBackgroundColor();
       stack.defer(async () => {
         await this.#emulationManager
@@ -1070,23 +1029,22 @@ export class CdpPage extends Page {
       clip = getIntersectionRect(clip, viewport);
     }
 
-    // We need to do these spreads because Firefox doesn't allow unknown options.
     const {data} = await this.#primaryTargetClient.send(
       'Page.captureScreenshot',
       {
         format: type,
-        ...(optimizeForSpeed ? {optimizeForSpeed} : {}),
+        optimizeForSpeed,
+        fromSurface,
         ...(quality !== undefined ? {quality: Math.round(quality)} : {}),
         ...(clip ? {clip: {...clip, scale: clip.scale ?? 1}} : {}),
-        ...(!fromSurface ? {fromSurface} : {}),
         captureBeyondViewport,
-      }
+      },
     );
     return data;
   }
 
   override async createPDFStream(
-    options: PDFOptions = {}
+    options: PDFOptions = {},
   ): Promise<ReadableStream<Uint8Array>> {
     const {timeout: ms = this._timeoutSettings.timeout()} = options;
     const {
@@ -1104,21 +1062,24 @@ export class CdpPage extends Page {
       omitBackground,
       tagged: generateTaggedPDF,
       outline: generateDocumentOutline,
+      waitForFonts,
     } = parsePDFOptions(options);
 
     if (omitBackground) {
       await this.#emulationManager.setTransparentBackgroundColor();
     }
 
-    await firstValueFrom(
-      from(
-        this.mainFrame()
-          .isolatedRealm()
-          .evaluate(() => {
-            return document.fonts.ready;
-          })
-      ).pipe(raceWith(timeout(ms)))
-    );
+    if (waitForFonts) {
+      await firstValueFrom(
+        from(
+          this.mainFrame()
+            .isolatedRealm()
+            .evaluate(() => {
+              return document.fonts.ready;
+            }),
+        ).pipe(raceWith(timeout(ms))),
+      );
+    }
 
     const printCommandPromise = this.#primaryTargetClient.send(
       'Page.printToPDF',
@@ -1140,11 +1101,11 @@ export class CdpPage extends Page {
         preferCSSPageSize,
         generateTaggedPDF,
         generateDocumentOutline,
-      }
+      },
     );
 
     const result = await firstValueFrom(
-      from(printCommandPromise).pipe(raceWith(timeout(ms)))
+      from(printCommandPromise).pipe(raceWith(timeout(ms))),
     );
 
     if (omitBackground) {
@@ -1154,25 +1115,26 @@ export class CdpPage extends Page {
     assert(result.stream, '`stream` is missing from `Page.printToPDF');
     return await getReadableFromProtocolStream(
       this.#primaryTargetClient,
-      result.stream
+      result.stream,
     );
   }
 
-  override async pdf(options: PDFOptions = {}): Promise<Buffer> {
+  override async pdf(options: PDFOptions = {}): Promise<Uint8Array> {
     const {path = undefined} = options;
     const readable = await this.createPDFStream(options);
-    const buffer = await getReadableAsBuffer(readable, path);
-    assert(buffer, 'Could not create buffer');
-    return buffer;
+    const typedArray = await getReadableAsTypedArray(readable, path);
+    assert(typedArray, 'Could not create typed array');
+    return typedArray;
   }
 
   override async close(
-    options: {runBeforeUnload?: boolean} = {runBeforeUnload: undefined}
+    options: {runBeforeUnload?: boolean} = {runBeforeUnload: undefined},
   ): Promise<void> {
+    using _guard = await this.browserContext().waitForScreenshotOperations();
     const connection = this.#primaryTargetClient.connection();
     assert(
       connection,
-      'Protocol error: Connection closed. Most likely the page has been closed.'
+      'Protocol error: Connection closed. Most likely the page has been closed.',
     );
     const runBeforeUnload = !!options.runBeforeUnload;
     if (runBeforeUnload) {
@@ -1212,12 +1174,12 @@ export class CdpPage extends Page {
    *   page.click('#connect-bluetooth'),
    * ]);
    * await devicePrompt.select(
-   *   await devicePrompt.waitForDevice(({name}) => name.includes('My Device'))
+   *   await devicePrompt.waitForDevice(({name}) => name.includes('My Device')),
    * );
    * ```
    */
   override async waitForDevicePrompt(
-    options: WaitTimeoutOptions = {}
+    options: WaitTimeoutOptions = {},
   ): Promise<DeviceRequestPrompt> {
     return await this.mainFrame().waitForDevicePrompt(options);
   }
@@ -1242,7 +1204,7 @@ const supportedMetrics = new Set<string>([
 /** @see https://w3c.github.io/webdriver-bidi/#rectangle-intersection */
 function getIntersectionRect(
   clip: Readonly<ScreenshotClip>,
-  viewport: Readonly<Protocol.DOM.Rect>
+  viewport: Readonly<Protocol.DOM.Rect>,
 ): ScreenshotClip {
   // Note these will already be normalized.
   const x = Math.max(clip.x, viewport.x);
@@ -1252,11 +1214,29 @@ function getIntersectionRect(
     y,
     width: Math.max(
       Math.min(clip.x + clip.width, viewport.x + viewport.width) - x,
-      0
+      0,
     ),
     height: Math.max(
       Math.min(clip.y + clip.height, viewport.y + viewport.height) - y,
-      0
+      0,
     ),
   };
 }
+
+export function convertCookiesPartitionKeyFromPuppeteerToCdp(
+  partitionKey: CookiePartitionKey | string | undefined,
+): Protocol.Network.CookiePartitionKey | undefined {
+  if (partitionKey === undefined) {
+    return undefined;
+  }
+  if (typeof partitionKey === 'string') {
+    return {
+      topLevelSite: partitionKey,
+      hasCrossSiteAncestor: false,
+    };
+  }
+  return {
+    topLevelSite: partitionKey.sourceOrigin,
+    hasCrossSiteAncestor: partitionKey.hasCrossSiteAncestor ?? false,
+  };
+}
